# -*- coding: utf-8 -*-

from __future__ import print_function, division, absolute_import

import os
import sys
import fnmatch
import ast
import doctest
import collections
import os.path as pathlib

from .log import logger
from .helpers import parse_git_config, trim_suffix

import nbformat

Module = collections.namedtuple('Module', ['name', 'try_', 'file', 'lineno'])


def parse_imports(package_root, ignores=None):
    """package_root must be a absolute path to package root,
    e.g. /path/to/pigar/pigar."""
    ignores = set(ignores) if ignores else set()
    ignores |= set([".hg", ".svn", ".git", "__pycache__"])
    ignored_paths = collections.defaultdict(set)
    for path in ignores:
        path = trim_suffix(path, "/")
        ignored_paths[pathlib.dirname(path)].add(pathlib.basename(path))

    imported_modules = []
    user_modules = set()

    for dirpath, dirnames, files in os.walk(package_root, followlinks=True):
        if dirpath in ignored_paths:
            dirnames[:] = [
                d for d in dirnames if d not in ignored_paths[dirpath]
            ]
        has_py = False
        for fn in files:
            fpath = pathlib.join(dirpath, fn)
            # C extension.
            if fn.endswith('.so'):
                has_py = True
                user_modules.add(fpath[:-3])
            # Normal Python file.
            elif fn.endswith('.py'):
                has_py = True
                user_modules.add(fpath[:-3])
            code = _read_code(fpath)
            if code:
                imported_modules.extend(parse_file_imports(fpath, code))
        if has_py:
            user_modules.add(trim_suffix(dirpath, "/"))
    return imported_modules, user_modules


def _read_code(fpath):
    if fpath.endswith(".ipynb"):
        nb = nbformat.read(fpath, as_version=4)
        code = ""
        for cell in nb.cells:
            if cell.cell_type == "code":
                code += cell.source + "\n"
        return code
    elif fpath.endswith(".py"):
        with open(fpath, 'rb') as f:
            return f.read()
    return None


def parse_file_imports(fpath, content):
    py_codes = collections.deque([(content, 1)])
    parser = ImportsParser(
        lambda code, lineno: py_codes.append((code, lineno))  # noqa
    )

    while py_codes:
        code, lineno = py_codes.popleft()
        try:
            parser.parse(code, fpath, lineno)
        except SyntaxError as e:
            # Ignore SyntaxError in Python code.
            logger.warn("parse %s:%d failed: %r", fpath, lineno, e)
    return parser.modules


class ImportsParser(object):
    def __init__(self, rawcode_callback=None):
        self._modules = []
        self._rawcode_callback = rawcode_callback

    def parse(self, content, fpath, lineno):
        parsed = ast.parse(content)
        self._fpath = fpath
        self._mods = fpath[:-3].split("/")
        self._lineno = lineno - 1
        self.visit(parsed)

    def _add_module(self, name, try_, lineno):
        self._modules.append(
            Module(name=name, try_=try_, file=self._fpath, lineno=lineno)
        )

    def _add_rawcode(self, code, lineno):
        if self._rawcode_callback:
            self._rawcode_callback(code, lineno)

    def visit_Import(self, node, try_=False):
        """As we know: `import a [as b]`."""
        lineno = node.lineno + self._lineno
        for alias in node.names:
            self._add_module(alias.name, try_, lineno)

    def visit_ImportFrom(self, node, try_=False):
        """
        As we know: `from a import b [as c]`. If node.level is not 0,
        import statement like this `from .a import b`.
        """
        mod_name = node.module
        level = node.level
        if mod_name is None:
            level -= 1
            mod_name = ""
        for alias in node.names:
            name = mod_name
            if level > 0 or mod_name == "":
                name = level*"." + mod_name + "." + alias.name
            lineno = node.lineno + self._lineno
            self._add_module(name, try_, lineno)

    def visit_TryExcept(self, node):
        """
        If modules which imported by `try except` and not found,
        maybe them come from other Python version.
        """
        for ipt in node.body:
            if ipt.__class__.__name__.startswith('Import'):
                method = 'visit_' + ipt.__class__.__name__
                getattr(self, method)(ipt, True)
        for handler in node.handlers:
            for ipt in handler.body:
                if ipt.__class__.__name__.startswith('Import'):
                    method = 'visit_' + ipt.__class__.__name__
                    getattr(self, method)(ipt, True)

    # For Python 3.3+
    visit_Try = visit_TryExcept

    def visit_Exec(self, node):
        """
        Check `expression` of `exec(expression[, globals[, locals]])`.
        **Just available in python 2.**
        """
        if hasattr(node.body, 's'):
            self._add_rawcode(node.body.s, node.lineno + self._lineno)
        # PR#13: https://github.com/damnever/pigar/pull/13
        # Sometimes exec statement may be called with tuple in Py2.7.6
        elif hasattr(node.body, 'elts') and len(
            node.body.elts
        ) >= 1 and hasattr(node.body.elts[0], 's'):
            self._add_rawcode(node.body.elts[0].s, node.lineno + self._lineno)

    def visit_Expr(self, node):
        """
        Check `expression` of `eval(expression[, globals[, locals]])`.
        Check `expression` of `exec(expression[, globals[, locals]])`
        in python 3.
        Check `name` of `__import__(name[, globals[, locals[,
        fromlist[, level]]]])`.
        Check `name` or `package` of `importlib.import_module(name,
        package=None)`.
        """
        # Built-in functions
        value = node.value
        if isinstance(value, ast.Call):
            lineno = node.lineno + self._lineno
            if hasattr(value.func, 'id'):
                if (
                    value.func.id == 'eval'
                    and hasattr(node.value.args[0], 's')
                ):
                    self._add_rawcode(node.value.args[0].s, lineno)
                # **`exec` function in Python 3.**
                elif (
                    value.func.id == 'exec'
                    and hasattr(node.value.args[0], 's')
                ):
                    self._add_rawcode(node.value.args[0].s, lineno)
                # `__import__` function.
                elif (
                    value.func.id == '__import__' and len(node.value.args) > 0
                    and hasattr(node.value.args[0], 's')
                ):
                    self._add_module(node.value.args[0].s, False, lineno)
            # `import_module` function.
            elif getattr(value.func, 'attr', '') == 'import_module':
                module = getattr(value.func, 'value', None)
                if (
                    module is not None
                    and getattr(module, 'id', '') == 'importlib'
                ):
                    args = node.value.args
                    arg_len = len(args)
                    if arg_len > 0 and hasattr(args[0], 's'):
                        name = args[0].s
                        if not name.startswith('.'):
                            self._add_module(name, False, lineno)
                        elif arg_len == 2 and hasattr(args[1], 's'):
                            self._add_module(args[1].s, False, lineno)

    def visit_FunctionDef(self, node):
        """
        Check docstring of function, if docstring is used for doctest.
        """
        docstring = self._parse_docstring(node)
        if docstring:
            self._add_rawcode(docstring, node.lineno + self._lineno + 2)

    def visit_ClassDef(self, node):
        """
        Check docstring of class, if docstring is used for doctest.
        """
        docstring = self._parse_docstring(node)
        if docstring:
            self._add_rawcode(docstring, node.lineno + self._lineno + 2)

    def visit(self, node):
        """Visit a node, no recursively."""
        for node in ast.walk(node):
            method = 'visit_' + node.__class__.__name__
            getattr(self, method, lambda x: x)(node)

    @staticmethod
    def _parse_docstring(node):
        """Extract code from docstring."""
        docstring = ast.get_docstring(node)
        if docstring:
            parser = doctest.DocTestParser()
            try:
                dt = parser.get_doctest(docstring, {}, None, None, None)
            except ValueError:
                # >>> 'abc'
                pass
            else:
                examples = dt.examples
                return '\n'.join([example.source for example in examples])
        return None

    @property
    def modules(self):
        return self._modules


def parse_installed_packages():
    """Get mapping for import top level name
    and install package name with version.
    """
    mapping = dict()

    for path in sys.path:
        if pathlib.isdir(path) and trim_suffix(path, '/').endswith(
            ('site-packages', 'dist-packages')
        ):
            mapping.update(_search_path(path))

    return mapping


def _search_path(path):
    mapping = dict()

    for file in os.listdir(path):
        # Install from PYPI.
        if fnmatch.fnmatch(file, '*-info'):
            top_level = os.path.join(path, file, 'top_level.txt')
            pkg_name, version = file.split('-')[:2]
            if version.endswith('dist'):
                version = version.rsplit('.', 1)[0]
            # Issue for ubuntu: sudo pip install xxx
            elif version.endswith('egg'):
                version = version.rsplit('.', 1)[0]
            mapping[pkg_name] = (pkg_name, version)
            if not os.path.isfile(top_level):
                continue
            with open(top_level, 'r') as f:
                for line in f:
                    mapping[line.strip()] = (pkg_name, version)

        # Install from local and available in GitHub.
        elif fnmatch.fnmatch(file, '*-link'):
            link = os.path.join(path, file)
            if not os.path.isfile(link):
                continue
            # Link path.
            with open(link, 'r') as f:
                for line in f:
                    line = line.strip()
                    if line != '.':
                        dev_dir = line
            if not dev_dir:
                continue
            # Egg info path.
            info_dir = [
                _file for _file in os.listdir(dev_dir)
                if _file.endswith('egg-info')
            ]
            if not info_dir:
                continue
            info_dir = info_dir[0]
            top_level = os.path.join(dev_dir, info_dir, 'top_level.txt')
            # Check whether it can be imported.
            if not os.path.isfile(top_level):
                continue

            # Check .git dir.
            git_path = os.path.join(dev_dir, '.git')
            if os.path.isdir(git_path):
                config = parse_git_config(git_path)
                url = config.get('remote "origin"', {}).get('url')
                if not url:
                    continue
                branch = 'branch "master"'
                if branch not in config:
                    for section in config:
                        if 'branch' in section:
                            branch = section
                            break
                if not branch:
                    continue
                branch = branch.split()[1][1:-1]

                pkg_name = info_dir.split('.egg')[0]
                git_url = 'git+{0}@{1}#egg={2}'.format(url, branch, pkg_name)
                with open(top_level, 'r') as f:
                    for line in f:
                        mapping[line.strip()] = ('-e', git_url)
    return mapping



# -*- coding: utf-8 -*-

from __future__ import print_function, division, absolute_import

import os
import sqlite3
import contextlib
try:  # py2
    from string import lowercase
except ImportError:  # py3
    from string import ascii_lowercase as lowercase

from .helpers import Dict, retry

# TODO(damnever): insert into db by default.
# (import_name, package_name)
_F_PACKAGES = {
    'yaml': 'PyYAML',
    'dogpile.cache': 'dogpile.cache',
    'dogpile.core': 'dogpile.core',
}


class Database(object):
    """Database store (top_level_name, package_name) piars.

    `top_level_name` is name can be imported from package,
    use `name` replace in db.
    `package_name` is name which be installed by pip, use
    `package` replace in db.

    Split table by `top_level_name` first letter, such as
    'table_a', `package_name` stored in `table_packages`.
    """
    _DB = os.path.join(os.path.dirname(__file__), '.db.sqlite3')
    _TABLE_PREFIX = 'table_{0}'
    _TABLE_OTHER_SUFFIX = 'lambda'
    _TABLE_PACKAGES = 'table_packages'

    def __init__(self, db=_DB):
        exist = os.path.isfile(db)
        self._db = db
        self._conn = None
        self._reconnect()
        if not exist:
            self._create_tables()

    def _reconnect(self):
        self.close()
        self._conn = sqlite3.connect(
            self._db, timeout=15
        )  # Avoid lock exception..

    def close(self):
        if self._conn is not None:
            self._conn.close()
            self._conn = None

    def insert_package_with_imports(self, pkgname, inames):
        package_table = self._package_table()
        sql = 'INSERT OR IGNORE INTO {0} (package) VALUES (?)'.format(
            package_table
        )
        sqls = [(sql, (pkgname, ))]
        sqltpl = '''INSERT OR IGNORE INTO {0} (name, pkgid) VALUES
        (?, (SELECT id from {1} WHERE package=?))'''
        for iname in inames:
            iname = iname or pkgname  # empty top_level.txt
            sql = sqltpl.format(self._name_table(iname[0]), package_table)
            sqls.append((sql, (iname, pkgname)))
        self.insert(sqls)

    @retry(sqlite3.OperationalError, count=3)
    def query_all(self, name):
        name_table = self._name_table(name[0])
        package_table = self._package_table()
        sql = '''SELECT {0}.name , {1}.package
        FROM {0} INNER JOIN {1} ON {0}.pkgid == {1}.id
        WHERE name=?'''.format(name_table, package_table)
        return self.query(sql, name)

    @retry(sqlite3.OperationalError, count=3)
    def query_package(self, package):
        package_table = self._package_table()
        if package:
            sql = 'SELECT * FROM {0} WHERE package=?'.format(package_table)
            rows = self.query(sql, package)
        else:
            sql = 'SELECT package FROM {0}'.format(package_table)
            rows = self.query(sql)
        if package:
            return rows[0] if rows else None
        else:
            return [r.package for r in rows]

    @retry(sqlite3.OperationalError, count=5)
    def insert(self, sqls):
        cursor = self._conn.cursor()
        conn = self._conn
        try:
            for (sql, params) in sqls:
                cursor.execute(sql, params)
            cursor.close()
            conn.commit()
        except sqlite3.OperationalError:
            cursor.close()
            try:
                conn.rollback()
            except Exception:
                pass
            self._reconnect()
            raise

    @retry(sqlite3.OperationalError, count=3)
    def query(self, sql, *parameters):
        cursor = self._conn.cursor()
        try:
            self._execute(cursor, sql, *parameters)
            names = [d[0] for d in cursor.description]
            return [Dict(zip(names, row)) for row in cursor]
        finally:
            cursor.close()

    def _execute(self, cursor, sql, *parameters):
        conn = self._conn
        try:
            result = cursor.execute(sql, parameters)
            conn.commit()
            return result
        except sqlite3.OperationalError:
            try:
                conn.rollback()
            except Exception:
                pass
            self._reconnect()
            raise

    def _name_table(
        self, initial, prefix=_TABLE_PREFIX, other=_TABLE_OTHER_SUFFIX
    ):
        initial = initial.lower()
        if initial not in lowercase:
            initial = other
        return prefix.format(initial)

    def _package_table(self, pkg_table=_TABLE_PACKAGES):
        return pkg_table

    def _create_tables(self, other=_TABLE_OTHER_SUFFIX):
        cursor = self._conn.cursor()
        try:
            # Create table `table_packages`.
            sql = '''CREATE TABLE IF NOT EXISTS {0} (
                id INTEGER PRIMARY KEY,  -- id will auto increment
                package VARCHAR NOT NULL UNIQUE
            )'''.format(self._package_table())
            self._execute(cursor, sql)

            # Create `table_[a-z]`.
            sql = '''CREATE TABLE IF NOT EXISTS {0} (
                id INTEGER PRIMARY KEY,
                name VARCHAR NOT NULL,
                pkgid INTEGER NOT NULL
                -- FOREIGN KEY(pkgid) REFERENCES packages(id)
            )'''
            for initial in (list(lowercase) + [other]):
                table = self._name_table(initial)
                self._execute(cursor, sql.format(table))
        finally:
            cursor.close()


@contextlib.contextmanager
def database():
    """A Database shortcut can auto close."""
    db = Database()
    yield db
    db.close()



# -*- coding: utf-8 -*-

from __future__ import print_function, division, absolute_import

import os
import os.path as pathlib
import sys
import unittest

from ..core import parse_packages, is_stdlib, _RequiredModules, _Locations


class ReqsTests(unittest.TestCase):
    def setUp(self):
        self._installed_packages = {
            'foo': ('Foo', '0.1.0'),
            'bar': ('Bar', '1.1.1'),
            'baz': ('Baz', '2.2.2'),
            'foobaz': ('FooBaz', '20151110'),
            'mod': ('Mod', '1.0.0'),
            'name': ('Name', '1.0.0'),
            'pkg': ('Pkg', '1.0.0'),
            'notebook': ('Notebook', '0.9.0'),
            'mainfoobar': ('min-foo-bar', '0.10.0rc0'),
        }
        self._path = os.path.abspath(
            pathlib.join(os.path.dirname(__file__), 'imports_example/')
        )
        self._module_infos = {}
        for k, v in {
            'foobar': ['example1.py: 7'],
            'FooBar': ['example1.py: 7'],
            'Foo': ['example1.py: 9'],
            'Bar': ['example1.py: 11'],
            'FooBaz': ['example1.py: 13'],
            'Baz': ['example1.py: 27'],
            'queue': ['example1.py: 33'],
            'Mod': ['example1.py: 41'],
            'Name': ['example1.py: 43'],
            'Pkg': ['example1.py: 45'],
            'Notebook': ['notebook.ipynb: 3'],
            'min-foo-bar': ['mainfoobar.py: 2'],
        }.items():
            paths = []
            for subp in v:
                paths.append(pathlib.join(self._path, subp))
            self._module_infos[k] = paths

        if sys.version_info[0] == 2:
            self._installed_packages.update({'foobar': ('FooBar', '3.6.9')})

    def tearDown(self):
        del self._installed_packages
        del self._path
        del self._module_infos

    @unittest.skipIf(sys.version_info[0] != 3, 'Not python 3.x')
    def test_py3_requirements(self):
        pv = {k: v for k, v in self._installed_packages.values()}
        pkgs, guess = self._parse_packages()

        self.assertListEqual(sorted(pkgs.keys()), sorted(pv.keys()))
        # Assume 'foobar' is Py3 builtin package, no need install.
        self.assertListEqual(sorted(guess.keys()), ['foobar'])
        self._check_require_pkgs(pkgs, pv)
        self._check_guess(guess, pv)

    @unittest.skipIf(sys.version_info[0] != 2, 'Not python 2.x')
    def test_py2_requirements(self):
        self._installed_packages.update({'foobar': ('FooBar', '3.3.3')})
        pv = {k: v for k, v in self._installed_packages.values()}
        pkgs, guess = self._parse_packages()

        self.assertListEqual(sorted(pkgs.keys()), sorted(pv.keys()))
        self.assertListEqual(guess.keys(), ['queue'])
        self._check_require_pkgs(pkgs, pv)
        self._check_guess(guess, pv)

    def _check_require_pkgs(self, pkgs, pv):
        for pkg, detail in pkgs.sorted_items():
            if pkg not in pv:
                self.fail('"{0}" not installed'.format(pkg))
            self.assertEqual(detail.version, pv[pkg])
            self.assertListEqual(
                sorted(detail.comments.sorted_items()),
                sorted(self._module_infos[pkg])
            )

    def _check_guess(self, mods, pv):
        for mod, locs in mods.items():
            self.assertListEqual(
                sorted(locs.sorted_items()), sorted(self._module_infos[mod])
            )

    def _parse_packages(self):
        return parse_packages(self._path, [], self._installed_packages)


class StdlibTest(unittest.TestCase):
    def test_stdlib(self):
        self.assertTrue(is_stdlib('os'))
        self.assertTrue(is_stdlib('sys'))


class LocationsTests(unittest.TestCase):
    def setUp(self):
        self._data = {
            'oo/xx.py': 33,
            'bar/baz.py': 10,
        }

    def test_add(self):
        loc = _Locations()
        for file, lineno in self._data.items():
            loc.add(file, lineno)

        for file in self._data:
            if file not in loc:
                self.fail('add "{0}" failed'.format(file))
            else:
                self.assertEqual(loc[file], [self._data[file]])

        loc.add('oo/xx.py', 2)
        self.assertListEqual(sorted(loc['oo/xx.py']), [2, 33])

    def test_extend(self):
        loc1 = _Locations()
        loc2 = _Locations()
        r1, r2 = self._data.items()
        loc1.add(r1[0], r1[1])
        loc2.add(r2[0], r2[1])

        loc1.extend(loc2)

        self.assertListEqual(
            sorted(loc1.items()),
            sorted([(k, [v]) for k, v in self._data.items()])
        )

    def test_sorted_items(self):
        loc = _Locations()
        for file, lineno in self._data.items():
            loc.add(file, lineno)

        target = ['{0}: {1}'.format(k, v) for k, v in self._data.items()]
        self.assertListEqual(loc.sorted_items(), sorted(target))


class RequiredModulesTest(unittest.TestCase):
    def setUp(self):
        loc1 = _Locations()
        loc2 = _Locations()
        loc1.add('oo/xx.py', 33)
        loc2.add('bar/baz.py', 10)
        self._data = {
            'pigar': ('9.9.9', loc1),
            'foo': ('3.3.3', loc2),
        }

    def test_add(self):
        rm = _RequiredModules()
        for pkg, (ver, loc) in self._data.items():
            rm.add_locs(pkg, ver, loc)

        for pkg in self._data:
            if pkg not in rm:
                self.fail('add "{0}" failed'.format(pkg))
            else:
                detail = rm[pkg]
                val = self._data[pkg]
                self.assertEqual(detail.version, val[0])
                self.assertIs(detail.comments, val[1])

        rm.add('pigar', '9.9.9', 'foobar.py', 2)
        self.assertListEqual(
            rm['pigar'].comments.sorted_items(),
            sorted(['oo/xx.py: 33', 'foobar.py: 2'])
        )



# -*- coding: utf-8 -*-

from __future__ import print_function, division, absolute_import

import collections
import codecs
import functools
import glob
import importlib
import imp
import os
import os.path as pathlib
import sys
try:
    from types import FileType  # py2
except ImportError:
    from io import IOBase as FileType  # py3

from .db import database
from .log import logger
from .helpers import (
    Color, lines_diff, print_table, parse_requirements, trim_prefix,
    trim_suffix
)
from .parser import parse_imports, parse_installed_packages
from .pypi import PKGS_URL, Downloader, Updater

from requests.exceptions import HTTPError

# FIXME: dirty workaround..
_special_packages = {
    "dogpile.cache": "dogpile.cache",
    "dogpile.core": "dogpile.core",
    "ruamel.yaml": "ruamel.yaml",
    "ruamel.ordereddict": "ruamel.ordereddict",
}


class RequirementsGenerator(object):
    def __init__(
        self,
        package_root,
        save_path,
        ignores=None,
        cmp_operator="==",
        ref_comments=False
    ):
        self._package_root = package_root
        self._save_path = save_path
        self._ignores = ignores
        self._cmp_operator = cmp_operator
        self._ref_comments = ref_comments
        self._installed_pkgs = None

    def __call__(self):
        self.generate()

    @property
    def installed_pkgs(self):
        if self._installed_pkgs is None:
            self._installed_pkgs = parse_installed_packages()
        return self._installed_pkgs

    def generate(self):
        packages, guess = parse_packages(
            self._package_root, self._ignores, self.installed_pkgs
        )

        answer = 'n'
        if guess:
            print(Color.RED('The following modules are not found yet:'))
            self._print_uncertain_modules(guess)
            sys.stdout.write(
                Color.RED(
                    (
                        'Some of them may be not installed in local '
                        'environment.\nTry to search PyPI for the '
                        'missing modules and filter'
                        ' some unnecessary modules? (y/[N]) '
                    )
                )
            )
            sys.stdout.flush()
            answer = sys.stdin.readline().strip().lower()

        in_pypi = None
        if answer in ('y', 'yes'):
            print(Color.BLUE('Checking modules on the PyPI...'))
            in_pypi = self._check_on_pypi(packages, guess)

        old = self._read_requirements()
        self._write_requirements(packages)
        new = self._read_requirements()
        self._print_diff(old, new)

        if in_pypi:
            for name in in_pypi:
                del guess[name]
        if guess and answer in ('y', 'yes'):
            print(Color.RED('These modules are not found:'))
            self._print_uncertain_modules(guess)
            print(Color.RED('Maybe or you need update database.'))

    def _check_on_pypi(self, packages, guess):
        in_pypi = set()
        for name, locs in guess.items():
            logger.info('Checking %s on the PyPI ...', name)
            downloader = Downloader()
            with database() as db:
                rows = db.query_all(name)
                pkgs = [row.package for row in rows]
                if pkgs:
                    in_pypi.add(name)
                for pkg in _best_matchs(name, pkgs):
                    try:
                        latest = downloader.download_package(pkg).version()
                        packages.add_locs(pkg, latest, locs)
                    except HTTPError as e:
                        logger.error('checking %s failed: %e', pkg, e)
        return in_pypi

    def _print_uncertain_modules(self, modules):
        for name, locs in modules.items():
            print(
                '  {0} referenced from:\n    {1}'.format(
                    Color.YELLOW(name), '\n    '.join(locs.sorted_items())
                )
            )

    def _read_requirements(self):
        if not pathlib.isfile(self._save_path):
            return
        with codecs.open(self._save_path, 'rb', 'utf-8') as f:
            return f.readlines()

    def _write_requirements(self, packages):
        print(
            Color.GREEN(
                'Writing requirements to "{0}"'.format(self._save_path)
            )
        )
        package_root_parent = pathlib.dirname(
            trim_suffix(self._package_root, "/")
        ) + "/"
        ref_comments = self._ref_comments
        cmp_operator = self._cmp_operator

        with open(self._save_path, 'w+') as f:
            f.write(
                '# Automatically generated by '
                'https://github.com/damnever/pigar.\n'
            )
            if not ref_comments:
                f.write('\n')
            for k, v in packages.sorted_items():
                if ref_comments:
                    f.write('\n')
                    f.write(
                        ''.join(
                            [
                                '# {0}\n'.format(
                                    trim_prefix(c, package_root_parent)
                                ) for c in v.comments.sorted_items()
                            ]
                        )
                    )
                if k == '-e':
                    f.write('{0} {1}\n'.format(k, v.version))
                elif v:
                    f.write('{0} {1} {2}\n'.format(k, cmp_operator, v.version))
                else:
                    f.write('{0}\n'.format(k))

    def _print_diff(self, old, new):
        if not old:
            return
        is_diff, diffs = lines_diff(old, new)
        msg = 'Requirements file has been overwritten, '
        if is_diff:
            msg += 'here is the difference:'
            print('{0}\n{1}'.format(Color.YELLOW(msg), ''.join(diffs)), end='')
        else:
            msg += 'no difference.'
            print(Color.YELLOW(msg))


def check_requirements_latest_versions(
    check_path,
    ignores=None,
    comparison_operator="==",
    ref_comments=False,
):
    logger.debug('Starting check requirements latest version ...')
    files = list()
    reqs = dict()
    pkg_versions = list()
    installed_pkgs = None
    # If no requirements file given, check in current directory.
    if pathlib.isdir(check_path):
        logger.debug('Searching file in "{0}" ...'.format(check_path))
        files.extend(glob.glob(pathlib.join(check_path, '*requirements.txt')))
        # If not found in directory, generate requirements.
        if not files:
            print(
                Color.YELLOW(
                    'Requirements file not found, '
                    'generate requirements ...'
                )
            )
            save_path = os.path.join(check_path, 'requirements.txt')
            rg = RequirementsGenerator(
                check_path, save_path, ignores, comparison_operator,
                ref_comments
            )
            rg()
            installed_pkgs = rg.installed_pkgs
            files.append(save_path)
    else:
        files.append(check_path)
    for fpath in files:
        reqs.update(parse_requirements(fpath))

    logger.debug('Checking requirements latest version ...')
    installed_pkgs = installed_pkgs or parse_installed_packages()
    installed_pkgs = {v[0]: v[1] for v in installed_pkgs.values()}
    downloader = Downloader()
    for pkg in reqs:
        current = reqs[pkg]
        # If no version specifies in requirements,
        # check in installed packages.
        if current == '' and pkg in installed_pkgs:
            current = installed_pkgs[pkg]
        logger.debug('Checking "{0}" latest version ...'.format(pkg))
        try:
            latest = downloader.download_package(pkg).version()
        except HTTPError as e:
            logger.error('checking %s failed: %e', pkg, e)
        pkg_versions.append((pkg, current, latest))

    logger.debug('Checking requirements latest version done.')
    print_table(pkg_versions)


def search_packages_by_names(names):
    """Search package information by names(`import XXX`).
    """
    downloader = Downloader()
    results = collections.defaultdict(list)
    not_found = list()

    installed_pkgs = parse_installed_packages()
    for name in names:
        logger.debug('Searching package name for "{0}" ...'.format(name))
        # If exists in local environment, do not check on the PyPI.
        if name in installed_pkgs:
            results[name].append(list(installed_pkgs[name]) + ['local'])
        # Check information on the PyPI.
        else:
            rows = None
            with database() as db:
                rows = db.query_all(name)
            if rows:
                for row in rows:
                    try:
                        version = downloader.download_package(row.package
                                                              ).version()
                        results[name].append((row.package, version, 'PyPI'))
                    except HTTPError as e:
                        logger.error('checking %s failed: %e', row.package, e)
            else:
                not_found.append(name)

    for name in results:
        print('Found package(s) for "{0}":'.format(Color.GREEN(name)))
        print_table(results[name], headers=['PACKAGE', 'VERSION', 'WHERE'])
    if not_found:
        msg = '"{0}" not found.\n'.format(Color.RED(', '.join(not_found)))
        msg += 'Maybe you need to update the database.'
        print(Color.YELLOW(msg))


def update_database():
    """Update database."""
    print(Color.GREEN('Starting update database ...'))
    print(Color.YELLOW('The process will take a long time!!!'))
    logger.info('Crawling "{0}" ...'.format(PKGS_URL))
    try:
        updater = Updater()
    except Exception:
        logger.error("Fail to fetch all packages: ", exc_info=True)
        print(Color.RED('Operation aborted'))
        return

    try:
        updater.run()
        updater.wait()
    except (KeyboardInterrupt, SystemExit):
        # FIXME(damnever): the fucking signal..
        updater.cancel()
        print(Color.BLUE('Operation canceled!'))
    else:
        print(Color.GREEN('Operation done!'))


def parse_packages(package_root, ignores=None, installed_pkgs=None):
    imported_modules, user_modules = parse_imports(package_root, ignores)
    installed_pkgs = installed_pkgs or parse_installed_packages()
    packages = _RequiredModules()
    guess = collections.defaultdict(_Locations)

    try_imports = set()
    for module in imported_modules:
        name = module.name
        if is_user_module(module, user_modules, package_root):
            logger.debug("ignore imports from user module: %s", name)
            continue
        if is_stdlib(name) or is_stdlib(name.split('.')[0]):
            logger.debug("ignore imports from stdlib: %s", name)
            continue
        names = []
        special_name = '.'.join(name.split('.')[:2])
        # Flask extension.
        if name.startswith('flask.ext.'):
            names.append('flask')
            names.append('flask_' + name.split('.')[2])
        # Special cases..
        elif special_name in _special_packages:
            names.append(_special_packages[special_name])
        # Other.
        elif '.' in name:
            names.append(name.split('.')[0])
        else:
            names.append(name)

        for name in names:
            if name in installed_pkgs:
                pkg_name, version = installed_pkgs[name]
                packages.add(pkg_name, version, module.file, module.lineno)
            else:
                guess[name].add(module.file, module.lineno)
            if module.try_:
                try_imports.add(name)

    names = []
    for name in guess:
        if name in try_imports:
            names.append(name)
    for name in names:
        del guess[name]
    return packages, guess


def _best_matchs(name, pkgs):
    # If imported name equals to package name.
    if name in pkgs:
        return [pkgs[pkgs.index(name)]]
    # If not, return all possible packages.
    return pkgs


def is_user_module(module, user_modules, package_root):
    name = module.name
    if name.startswith("."):
        return True
    parts = name.split(".")
    cur_mod_path = module.file[:-3]
    dir_path_parts = pathlib.dirname(module.file).split("/")
    nparts = len(dir_path_parts)
    for i in range(0, nparts):
        i = -i if i > 0 else nparts
        dir_path = "/".join(dir_path_parts[:i])
        if dir_path == "":
            dir_path = "/"
        if dir_path not in user_modules:
            break
        mod_paths = [pathlib.join(dir_path, "/".join(parts))]
        if len(dir_path_parts[:i]) > 0 and dir_path_parts[:i][-1] == parts[0]:
            mod_paths.append(dir_path)
        for mod_path in mod_paths:
            # FIXME(damnever): ignore the current file?
            if mod_path == cur_mod_path:
                continue
            if mod_path in user_modules:
                return True
    return False


def _checked_cache(func):
    checked = dict()

    @functools.wraps(func)
    def _wrapper(name):
        if name not in checked:
            checked[name] = func(name)
        return checked[name]

    return _wrapper


@_checked_cache
def is_stdlib(name):
    """Check whether it is stdlib module."""
    exist = True
    module_info = ('', '', '')
    try:
        module_info = imp.find_module(name)
    except ImportError:
        try:
            # __import__(name)
            importlib.import_module(name)
            module_info = imp.find_module(name)
            sys.modules.pop(name)
        except ImportError:
            exist = False
    # Testcase: ResourceWarning
    if isinstance(module_info[0], FileType):
        module_info[0].close()
    mpath = module_info[1]
    if exist and (
        mpath is not None and (
            'site-packages' in mpath or 'dist-packages' in mpath or
            ('bin/' in mpath and mpath.endswith('.py'))
        )
    ):
        exist = False
    return exist


class _RequiredModules(dict):

    _Detail = collections.namedtuple('Detail', ['version', 'comments'])

    def __init__(self):
        super(_RequiredModules, self).__init__()
        self._sorted = None

    def add_locs(self, package, version, locations):
        if package in self:
            self[package].comments.extend(locations)
        else:
            self[package] = self._Detail(version, locations)

    def add(self, package, version, file, lineno):
        if package in self:
            self[package].comments.add(file, lineno)
        else:
            loc = _Locations()
            loc.add(file, lineno)
            self[package] = self._Detail(version, loc)

    def sorted_items(self):
        if self._sorted is None:
            self._sorted = sorted(self.items())
        return self._sorted

    def remove(self, *names):
        for name in names:
            if name in self:
                self.pop(name)
        self._sorted = None


class _Locations(dict):
    """_Locations store code locations(file, linenos)."""
    def __init__(self):
        super(_Locations, self).__init__()
        self._sorted = None

    def add(self, file, lineno):
        if file in self and lineno not in self[file]:
            self[file].append(lineno)
        else:
            self[file] = [lineno]

    def extend(self, obj):
        for file, linenos in obj.items():
            for lineno in linenos:
                self.add(file, lineno)

    def sorted_items(self):
        if self._sorted is None:
            self._sorted = [
                '{0}: {1}'.format(f, ','.join([str(n) for n in sorted(ls)]))
                for f, ls in sorted(self.items())
            ]
        return self._sorted



# -*- coding: utf-8 -*-

from __future__ import print_function, division, absolute_import

import unittest
import os

from ..pypi import _parse_indexed_packages


class ExtractHtmlTest(unittest.TestCase):
    def test_extract_names(self):
        path = os.path.join(
            os.path.dirname(__file__), './fake_simple_html.txt'
        )
        names = []
        with open(path) as f:
            _parse_indexed_packages(f.read(), names.append)
        self.assertListEqual(names, 'a b c d e f g'.split())



# -*- coding: utf-8 -*-

from __future__ import print_function, division, absolute_import

import unittest
import os
import tempfile

from ..db import Database


class DBTests(unittest.TestCase):

    def setUp(self):
        self._db = os.path.join(tempfile.gettempdir(), 'pigar_test.db')
        self._conn = Database(db=self._db)

    def tearDown(self):
        self._conn.close
        os.remove(self._db)

    def test_db(self):
        self._conn.insert_package_with_imports('pigar', [])
        row = self._conn.query_package('pigar')
        self.assertEqual(row.id, 1)
        self._conn.insert_package_with_imports('pigar', ['pigar'])
        row = self._conn.query_all('pigar')
        self.assertDictEqual(dict(row[0]),
                             {'name': 'pigar', 'package': 'pigar'})
        rows = self._conn.query_package(None)
        self.assertListEqual(rows, ['pigar'])
